from durin import *
import time


if __name__ == "__main__":

    # We start a connection to the robot
    # and can now read from and write to the robot via the variable "durin"
    # Notice the UI class, which differs from the (more efficient) standalone Durin interface
    with DurinUI("durin1.local") as durin:
        # Loop until the user quits
        is_running = True

        sensor_frequencies = (["Imu", 20],
                              ["Position", 20],
                              ["SystemStatus", 1000],
                              ["Uwb", 20],
                              ["Tof", 20],
                              )

        
        for sensor in sensor_frequencies:
            durin(SetSensorPeriod(sensor[0],sensor[1]))


        while is_running:

            # Read a value from durin
            # - obs = Robot sensor observations
            # - dvs = Robot DVS data (if any)
            # - cmd = Robot responses to commands
            (obs, dvs, cmd) = durin.read()

            # We can now update our display with the observations
            durin.render_sensors(obs)

            # Read user input and quit, if asked
            is_running = durin.read_user_input()



import numpy as np


class RingBuffer(object):
    def __init__(self, array: np.ndarray, counter: int = 0):
        self.size = len(array)
        self.buffer = array
        self.counter = counter

    def append(self, data):
        self.buffer[self.counter] = data
        self.counter += 1
        self.counter = self.counter % self.size
        return self.buffer

    def get_oldest(self):
        return self.buffer[self.counter]

    def get_newest(self):
        return self.buffer[(self.counter - 1 ) % self.size]




import time

from durin import *

if __name__ == "__main__":

    # We start a connection to the robot
    # and can now read from and write to the robot via the variable "durin"
    #with Durin("durin5.local", enable_dvs=True) as durin:
    with Durin("172.16.222.200", enable_control=False, enable_dvs=True) as durin:

        # Loop forever
        while True:

            # Read a value from durin
            # - obs = Robot sensor observations
            # - dvs = Robot DVS data (if any)
            # - cmd = Robot responses to commands
            (obs, dvs, cmd) = durin.read()
            if dvs is not None:
                print(dvs.sum())

            # Move following a square-shaped trajectory
            durin(Move(100, 0, 0))
            time.sleep(1)
            durin(Move(0, 100, 0))
            time.sleep(1)
            durin(Move(-100, 0, 0))
            time.sleep(1)
            durin(Move(0, -100, 0))
            time.sleep(1)



import ipaddress
import logging
from queue import Empty, Full
import socket
import multiprocessing
from typing import ByteString, Optional
from durin import io
from durin.actuator import StreamOn
from durin.io.runnable import RunnableConsumer, RunnableProducer


def get_ip(ip):
    # Thanks to https://stackoverflow.com/a/28950776/999865
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        s.connect((ip, 1))
        ip_guess = s.getsockname()[0]
    except Exception:
        ip_guess = "127.0.0.1"
    finally:
        s.close()
    return ip_guess

def checksum(bs):
    checksum = bs[0]
    for b in bs[1:]:
        checksum = checksum ^ b
    return checksum


class TCPProducer(RunnableProducer):

    def produce(self, sock):
        try:
            header = sock.recv(3)
            size = int.from_bytes(header[1:], 'little')
            size = size & (0b0000111111111111) # Zero out meta-bits
            bs = sock.recv(size)
            check_actual = int.from_bytes(sock.recv(1), 'little')
            check_bytes = checksum(header + bs)
            if (check_actual == check_bytes):
                return bs
            else:
                logging.warning(f"Checksum failed: {check_actual} != {check_bytes}")
                return None

        except BlockingIOError:
            return None


class TCPConsumer(RunnableConsumer):

    def consume(self, event, sock):
        bs = b'*' + len(event).to_bytes(2, "little") + event
        bs = bs + checksum(bs).to_bytes(1, 'little')
        sock.send(bs)


class TCPLink:
    """ """

    def __init__(
        self,
        host: str,
        port: str,
        buffer_size_send: int = 50,
        buffer_size_receive: int = 2000,
    ):
        address = (host, int(port))
        context = multiprocessing.get_context("spawn")
        # Buffer towards Durin
        self.buffer_send = context.Queue(buffer_size_send)
        # Buffer receiving replies
        self.buffer_receive = context.Queue(buffer_size_receive)

        # Create socket
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self.sock.connect(address)
            self.sock.setblocking(False)
        except (ConnectionRefusedError, OSError) as e:
            raise ConnectionRefusedError(f"Cannot reach Durin at {address}")

        # Create producer/consumer
        self.sender = TCPConsumer(self.buffer_send, self.sock)
        self.receiver = TCPProducer(self.buffer_receive, self.sock)

    def start(self):
        self.sender.start()
        self.receiver.start()

    def stop(self):
        self.sock.close()
        self.buffer_receive.close()
        self.buffer_send.close()
        self.sender.stop()
        self.receiver.stop()
        logging.debug(f"TCP control communication stopped")

    def send(self, command: ByteString, timeout: float) -> None:
        try:
            self.buffer_send.put(command, block=False, timeout=timeout)
        except Full:
            print("hej")
            return None

    def read(self) -> Optional[ByteString]:
        try:
            return self.buffer_receive.get(block=False)
        except Empty:
            return None


class UDPLink(RunnableProducer):
    """
    An UDPBuffer that silently buffers messages received over UDP into a queue.
    """

    def __init__(
        self, host: str, ip: str, packet_size: int = 512, buffer_size: int = 100
    ):
        self.buffer_size = buffer_size
        self.packet_size = packet_size
        context = multiprocessing.get_context("spawn")
        self.buffer = context.Queue(self.buffer_size)
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        address = (host, ip)
        self.sock.bind(address)
        self.sock.setblocking(0)
        logging.debug(f"UDP control receiving on {address}")

        super().__init__(self.buffer, self.sock)

    def produce(self, sock):
        try:
            buffer = sock.recv(self.packet_size)
            size = int.from_bytes(buffer[1:3], "little")
            size = size & (0b0000111111111111) # Zero out meta-bits
            check_actual = int.from_bytes(buffer[3 + size : 4 + size], 'little')
            check_bytes = checksum(buffer[:3 + size])
            if (int(check_actual) == check_bytes):
                msg = io.decode(buffer[3 : 3 + size])
                return msg
            else:
                logging.warning(f"Checksum failed: {check_actual} != {check_bytes}")
                return None
        except BlockingIOError:
            return None

    def stop(self):
        super().stop()
        self.sock.close()
        self.buffer.close()
        logging.debug(f"UDP communication stopped")


# class DVSClient(TCPLink):
#     def __init__(self, host: str, port: int) -> None:
#         self.sock = None
#         self.address = (host, port)

#     def _init_connection(self):
#         self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#         self.sock.connect(self.address)
#         self.sock.setblocking(0)
#         logging.debug(f"UDP DVS communication sending to {self.address}")

#     def _send_message(self, message: bytes):
#         try:
#             self.sock.send(message)
#         except (BrokenPipeError, AttributeError) as e:
#             try:
#                 self._init_connection()
#             except ConnectionRefusedError as e:
#                 raise ConnectionRefusedError(
#                     f"Could not connect to DVS controller at {self.address}"
#                 )
#             self._send_message(message)

#     def start_stream(self, host: str, port: int):
#         # cmd = StreamOn(host, port, 1)
#         # self._send_message(cmd.encode())
#         data = bytearray([0] * 7)
#         data[0] = 0x0
#         data[1:5] = int(ipaddress.ip_address(host)).to_bytes(4, "little")
#         data[5:7] = port.to_bytes(2, "little")
#         self._send_message(data)

#     def stop_stream(self):
#         if self.sock is not None:
#             self._send_message(bytes([1]))
#             self.sock.close()
#             self.sock = None



from abc import abstractmethod
from dataclasses import dataclass
from typing import ByteString, TypeVar
from pathlib import Path
from durin.io import SENSORS
from . import schema


def _wrap_base(message, field):
    base = schema.DurinBase.new_message()
    setattr(base, field, message)
    packed = base.to_bytes_packed()
    return packed


@dataclass
class Command:
    pass

    @abstractmethod
    def encode() -> ByteString:
        pass


class PowerOff(Command):
    def __init__(self):
        self.cmd_id = 1

    def encode(self):
        return _wrap_base(schema.PowerOff.new_message(), "powerOff")


class SetLed(Command):
    def __init__(self, red: int, green: int, blue: int):
        self.red = red
        self.green = green
        self.blue = blue

    def encode(self):
        return _wrap_base(
            schema.SetLed.new_message(ledR=self.red, ledG=self.green, ledB=self.blue),
            "setLed",
        )


class Move(Command):
    def __init__(self, vel_x: int, vel_y: int, rot: int):
        """
        Moves Durin

        Arguments:
            vel_x (int): Velocity in the x axis
            vel_y (int): Velocity in the y axis
            rot (int): Degrees per second
        """
        self.cmd_id = 2
        self.vel_x = int(vel_x)
        self.vel_y = int(vel_y)
        self.rot = int(rot)

    def encode(self):
        message = schema.SetRobotVelocity.new_message(
            velocityXMms=self.vel_x, velocityYMms=self.vel_y, rotationDegs=self.rot
        )
        return _wrap_base(message, "setRobotVelocity")

    def __repr__(self) -> str:
        return f"Move({self.vel_x}, {self.vel_y}, {-self.rot})"


class MoveWheels(Command):
    """
    Moves individual wheels on Durin. All units are in millimeters/second

    Arguments:
        front_left (float): Front left wheel
        front_right (float): Front right
        back_left (float): Back left wheel
        back_right (float): Back right wheel
    """

    def __init__(self, front_left, front_right, back_left, back_right):
        MAX_VALUE = 10000
        self.cmd_id = 3
        self.front_left = min(MAX_VALUE, int(front_left))
        self.front_right = min(MAX_VALUE, int(front_right))
        self.back_left = min(MAX_VALUE, int(back_left))
        self.back_right = min(MAX_VALUE, int(back_right))

    def encode(self):
        return _wrap_base(
            schema.SetWheelVelocity.new_message(
                wheelFrontLeftMms=self.front_left,
                wheelFrontRightMms=self.front_right,
                wheelBackLeftMms=self.back_left,
                wheelBackRightMms=self.back_right,
            ), "setWheelVelocity"
        )

    def __repr__(self) -> str:
        return f"MoveWheels({self.front_left}, {self.front_right}, {self.back_left}, {self.back_right})"


# class PollSensor(Command):
#     def __init__(self, sensor_id):
#         self.cmd_id = 17
#         self.sensor_id = int(sensor_id)

#     def encode(self):
#         data = bytearray([0] * 2)
#         data[0] = self.cmd_id
#         data[1:2] = self.sensor_id.to_bytes(
#             1, "little"
#         )  # integer (uint8) little endian

#         return data

class SetSensorPeriod(Command):

    _SUPPORTED_SENSORS = ["Tof", "Imu", "SystemStatus", "Position", "Uwb"]

    def __init__(self, sensor: str, period: int):
        """
        Sets how often a given sensor should report to the client in ms.
        """
        if not sensor in self._SUPPORTED_SENSORS:
            raise ValueError(f"Sensor '{sensor}' unsupported. Please choose from {self._SUPPORTED_SENSORS}")
        
        
        self.sensor = sensor
        self.period = period

    def encode(self):
        message_type = f"Set" + self.sensor + "StreamPeriod"

        message = getattr(schema, message_type).new_message()
        message.periodMs = self.period

        return _wrap_base(message, message_type[0].lower() + message_type[1:])


class StreamOn(Command):
    def __init__(self, host, port):
        self.cmd_id = 18
        self.host = host
        self.port = port

    def encode(self):
        enable_message = schema.EnableStreaming.new_message()
        udpOnly = enable_message.destination.init("udpOnly")
        udpOnly.ip = list(map(lambda x: int(x), self.host.split(".")))
        udpOnly.port = self.port
        return _wrap_base(enable_message, "enableStreaming")


class StreamOff(Command):
    def __init__(self):
        self.cmd_id = 19

    def encode(self):
        return _wrap_base(schema.DisableStreaming.new_message(), "disableStreaming")



import norse
import torch

from durin import *

if __name__ == "__main__":

    # First, we create a small simple Braitenberg vehicle with Norse
    # - https://en.wikipedia.org/wiki/Braitenberg_vehicle
    network = norse.torch.LICell(p=norse.torch.LIParameters(tau_mem_inv=100))
    # Prepare an empty (default) neuron state for later updating
    network_state = None

    # We start a connection to the robot
    # and can now read from and write to the robot via the variable "durin"
    # Notice the UI class, which differs from the (more efficient) standalone Durin interface
    with DurinUI("durin5.local") as durin:
        # Loop until the user quits
        is_running = True
        while is_running:
            # Read a value from durin
            # - obs = Robot sensor observations
            # - dvs = Robot DVS data (if any)
            # - cmd = Robot responses to commands
            (obs, dvs, cmd) = durin.read()

            # We can now update our display with the observations
            durin.render_sensors(obs)

            # Read user input and quit, if asked (but without allowing the user to move)
            is_running = durin.read_user_input(allow_movement=False)

            # We can now update our neural network and send the actuator signal to Durin
            # - Take the center pixel of the 2nd sensor in a (8, 8) matrix
            left_tof = obs.tof[1, 3, 3]
            # - Take the center pixel of the 7th sensor in a (8, 8) matrix
            right_tof = obs.tof[6, 3, 3]

            # We normalize the ToF information to [0;1] and flip it numerically
            # (so, close = high values, far = low values)
            input_tensor = torch.tanh(torch.tensor([left_tof, right_tof]) / 1e3) * -1 + 1

            # We then run the network (in inference mode)
            with torch.inference_mode():
                # Update the network with the input and read the output
                output, network_state = network(input_tensor, network_state)
                output = output * 100 # Scale the output to fit motor range [0;500]
                left, right = output  # Separate left and right outputs
                command = MoveWheels(left, right, right, left)
                durin(command)  # Send the command to Durin




import queue

from durin import io
from durin.io.command import *
from durin.io.network import TCPLink
from durin.controller import *

class DurinActuator:
    def __init__(self, tcp_link: TCPLink):
        self.tcp_link = tcp_link

    def __call__(self, action: Command, timeout: float = 0.05):
        command_bytes = action.encode()
        try:
            self.tcp_link.send(command_bytes, timeout=timeout)
        except queue.Full:
            pass

        return None

    def read(self):
        reply = self.tcp_link.read()
        if reply is not None:
            msg = io.decode(reply)
            return msg

        else:
            return None

    def start(self):
        self.tcp_link.start()

    def stop(self):
        self.tcp_link.stop()



from abc import abstractmethod
import multiprocessing
from multiprocessing.sharedctypes import Value
import queue
import time
from typing import Any


class Runnable:
    def __init__(self, *args, context: str = "spawn", **kwargs):
        context = multiprocessing.get_context(context)
        self.event = context.Event()
        self.thread = context.Process(target=self._run_thread, args=(self.event, *args, *kwargs.values()))

    @abstractmethod
    def run(self, *args):
        pass

    def _run_thread(self, *args):
        event = args[0]
        try:
            while not event.wait(0):
                self.run(*args[1:])
        except ValueError:
            pass
        except KeyboardInterrupt:
            pass

    def start(self):
        self.thread.start()

    def stop(self):
        self.event.set()
        if self.thread.is_alive():
            self.thread.terminate()


class RunnableConsumer(Runnable):
    def __init__(self, queue, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.queue = queue

    @abstractmethod
    def consume(self, item, *args):
        pass

    def run(self, *args):
        try:
            item = self.queue.get(block=False)
            if item is not None:
                self.consume(item, *args)
        except queue.Empty:
            pass


class RunnableProducer(Runnable):
    def __init__(self, queue, args, blocking: bool = False, **kwargs):
        super().__init__(args, **kwargs)
        self.queue = queue
        self.blocking = blocking

    @abstractmethod
    def produce(self, *args) -> Any:
        pass

    def run(self, *args):
        try:
            value = self.produce(*args)
            if value is not None:
                self.queue.put(value, block=self.blocking)
        except queue.Full:
            pass



import argparse

from durin.cli import run_cli
from durin.durin import Durin

if __name__ == "__main__":
    parser = argparse.ArgumentParser("Durin control interface")
    parser.add_argument("host", type=str, help="IP address to durin")
    parser.add_argument(
        "--port",
        type=int,
        default="4000",
        help="Port to use on Durin. Defaults to 4000",
    )
    args = parser.parse_args()
    durin = Durin(args.host, args.port)
    run_cli(durin)



import logging
from typing import Optional

from durin.actuator import StreamOn
from durin.actuator import DurinActuator
from durin.io.network import TCPLink, UDPLink
from durin.sensor import DurinSensor
import durin.io

DURIN_CONTROLLER_PORT_TCP = 1337
DURIN_CONTROLLER_PORT_UDP = 1338

DURIN_DVS_PORT_TCP = 2300


class Durin:
    """
    Interface to the Durin robot.
    This class first establishes connection to the controller and DVS system on board the robot.
    Afterwards, commands can be sent to the robot and sensory data can be `.read` from the robot.

    Note that the sensor observation is described in the sensor.Observation class.
    The DVS output is a 640x480 PyTorch tensor (if the aestream dependency is installed).

    Example:

    >>> with Durin(10.0.0.1) as durin:
    >>>   while True:
    >>>     observation, dvs = durin.read()
    >>>     ...
    >>>     durin(Move(x, y, rot))

    Arguments:
        durin_ip (str): The IPv4 address of the Durin microcontroller
        device (str): The PyTorch device to use for storing tensors. Defaults to cpu
        stream_command (Optional[StreamOn]): The command sent to the Durin microcontroller upon start. Can be customized, but uses sensible values by default.
        sensor_frequency (int): The update frequency of the sensor information in ms. Defaults to 15ms.
        enable_control (bool): Enables Durin sensors and control. Useful for testing without sending commands. Defaults to True.
        enable_dvs (bool): Enables Durin DVS data. Requires the aestream dependency AND a mounted DVS camera on Durin. Defaults to False.
    """

    def __init__(
        self,
        durin_ip: str,
        device: str = "cpu",
        stream_command: Optional[StreamOn] = None,
        sensor_frequency: int = 15,
        enable_control: bool = True,
        enable_dvs: bool = False,
    ):
        if stream_command is not None:
            self.stream_command_control = stream_command
            self.stream_command_dvs = StreamOn(stream_command.host, DURIN_DVS_PORT_TCP + 1)
        else:
            response_ip = durin.io.network.get_ip(durin_ip)
            self.stream_command_control = StreamOn(
                response_ip, DURIN_CONTROLLER_PORT_UDP
            )
            self.stream_command_dvs = StreamOn(response_ip, DURIN_DVS_PORT_TCP + 1)

        # Controller
        self.enable_control = enable_control
        if self.enable_control:
            tcp_link = TCPLink(durin_ip, DURIN_CONTROLLER_PORT_TCP)
            udp_link = UDPLink(self.stream_command_control.host, self.stream_command_control.port)
            self.sensor = DurinSensor(udp_link)
            self.actuator = DurinActuator(tcp_link)
            
        # DVS
        self.enable_dvs = enable_dvs
        if self.enable_dvs:
            import durin.dvs as dvs

            self.dvs_client = TCPLink(durin_ip, DURIN_DVS_PORT_TCP)
            self.dvs = dvs.DVSSensor((640, 480), device, DURIN_DVS_PORT_TCP + 1)
        else:
            logging.debug("DVS output disabled")

    def __enter__(self):
        # Controller
        if self.enable_control:
            self.sensor.start()
            self.actuator.start()
            # Start streaming
            self(self.stream_command_control)

            logging.debug(
                f"Durin Controller receiving on {self.stream_command_control.host}:{self.stream_command_control.port}"
            )

        # DVS
        if self.enable_dvs:
            self.dvs.start_stream()
            self.dvs_client.start()
            self.dvs_client.send(self.stream_command_dvs.encode(), 1000)
            logging.debug(
                f"Durin DVS receiving on {self.stream_command_dvs.host}:{self.stream_command_dvs.port + 1}"
            )

        return self

    def __exit__(self, e, b, t):
        if self.enable_control:
            self.sensor.stop()
            self.actuator.stop()
        if self.enable_dvs:
            self.dvs_client.stop()
            self.dvs.stop_stream()

    def __call__(self, command):
        if self.enable_control:
            return self.actuator(command)
        else:
            logging.warn(f"Command {command} sent, but Durin controls are disabled")

    def update_frequency(self) -> float:
        return 1 / self.sensor.freq / 6

    def read(self):
        """
        Retrieves sensor data from Durin

        Returns:
            Tuple of Observation, DVS tensors, and Command reply
        """
        dvs = self.dvs.read() if self.enable_dvs else None
        obs = self.sensor.read() if self.enable_control else None
        cmd = self.actuator.read() if self.enable_control else None
        return (obs, dvs, cmd)



import argparse
import os
import sys
import numpy as np
import functools
import time
import curses

from durin.actuator import Move

from durin.actuator import *


def parse(command: str, actuator: DurinActuator) -> str:
    try:
        out = eval(command)
        if isinstance(out, Command):
            reply = actuator(out)
            if reply is not None:
                return reply
            else:
                return "N/A"
        else:
            return f"Unknown command {out}"
    except Exception as e:
        return str(e)


def run_cli(actuator, stdin=sys.stdin.fileno()):
    # TODO: Print list of commands

    try:
        fd = os.fdopen(stdin)
        screen = curses.initscr()
        curses.noecho()
        curses.cbreak()

        initmsg = "Durin robot control environment\nAvailable commands:\n\tMove(x,y,a)"
        screen.addstr(0, 0, initmsg)

        vs = ""
        i = 3
        linestart = "> "
        while True:
            # Refresh
            screen.addstr(i, 2, f"{linestart}{vs} ")
            screen.addstr(i, 2, f"{linestart}{vs}")
            screen.refresh()

            key = screen.getch()

            if key == curses.KEY_ENTER or int(key) == 10:
                if len(vs) > 0:
                    val = parse(vs, actuator)
                    vs = ""
                    screen.addstr(i + 1, 0, str(val))
                    i += 2
            elif int(key) == 65:  # Up
                actuator(Move(0, 500, 0))
            elif int(key) == 66:  # Down
                actuator(Move(0, -500, 0))
            elif int(key) == 68:  # Left
                actuator(Move(-500, 0, 0))
            elif int(key) == 67:  # Right
                actuator(Move(500, 0, 0))
            elif key == curses.KEY_BACKSPACE or int(key) == 127:
                vs = vs[:-1]
                actuator(Move(0, 0, 0))
            else:
                vs = vs + chr(key)
    except (Exception, KeyboardInterrupt) as e:
        pass
    finally:
        curses.nocbreak()
        curses.echo()
        curses.endwin()


#### Figure out:


def parse_args():
    parser = argparse.ArgumentParser(description="Connect to Durin and exert control")

    parser.add_argument(
        "mode", type=str, help="Operating Mode: 'cli' or 'brain'", default="cli"
    )
    parser.add_argument(
        "--host", type=str, help="Durin's IP address", default="127.0.0.1"
    )
    parser.add_argument("--tcp", type=str, help="Durin's TCP port", default=2300)

    return parser.parse_args()


def show_content(filepath):
    with open(filepath) as fp:
        line = fp.readline()
        cnt = 1
        while line:
            print(f"cmd {line.strip()}")
            line = fp.readline()
            cnt += 1


def parse_line(cli_in):
    cmd_id = 0
    command = ""
    arg_nb = 0
    arg_array = []
    isvalid = False

    # List of available commands
    filepath = "commands.txt"

    # Ignore 'enter'
    if len(cli_in) > 0:

        # Break line into 'words' (i.e. command + arguments) and count arguments
        cli_in = cli_in.split()
        command = cli_in[0]
        arg_nb = len(cli_in) - 1
        arg_array = cli_in[1 : arg_nb + 1]

        # Check if line command matches any available command
        if command == "list":
            show_content(filepath)
        else:
            with open(filepath) as fp:
                while True:
                    line = fp.readline()
                    if not line:
                        break
                    line = line.split()

                    # Check if requested command matches current available one
                    if command == line[1]:
                        # Check if #args makes sense
                        if len(line) == 2 + len(arg_array):
                            cmd_id = line[0]
                            isvalid = True
                            break
            if cmd_id == 0:
                cli_in = []
                print("Wrong Command/Arguments ...\n")

    arg_array = np.array(arg_array, dtype=float)

    return command, arg_array, isvalid


if __name__ == "__main__":

    run_cli(lambda x: None)



import time

import torch

from durin import *


def cmd(x):
    x = torch.sin(torch.as_tensor(x)) * 200
    return Move(0, x, 0)


with Durin("172.16.223.95") as durin:

    tensors = []
    time_start = time.time()

    while True:
        (obs, tensor, cmd) = durin.read()
        tensors.append(tensor)

        # Pause to loop at 200Hz
        time.sleep(0.005)

        time_now = time.time()
        diff = time_now - time_start
        durin(cmd(diff))
        if time_now - time_start >= 10:
            break

    video = torch.stack(tensors)
    torch.save(video, "moving.dat")



from abc import ABC, abstractmethod
import logging
import multiprocessing
import time
from typing import Generic, NamedTuple, Tuple, TypeVar

import numpy as np
from durin.io import SENSORS

from durin.io.network import UDPLink
from durin.io.runnable import RunnableConsumer
from durin.io.ringbuffer import RingBuffer

T = TypeVar("T")


class Observation(NamedTuple):
    tof: np.ndarray = np.zeros((8, 8, 8))
    tof_frequency: float = 0
    charge: float = 0
    voltage: float = 0
    imu: np.ndarray = np.zeros((3, 3))
    uwb: np.ndarray = np.zeros((10, 2))
    position: np.ndarray = np.zeros((1, 3))
    imu_frequency: float = 0
    uwb_frequency: float = 0
    status_frequency: float = 0

    def __repr__(self) -> str:
        tof_str = " ".join([f"{x.mean():.0f}±{x.std():.0f}" for x in self.tof])
        return f"Durin {self.charge}%\n\tIMU: {self.imu}\n\tTOF: {tof_str}"


class Sensor(ABC, Generic[T]):
    @abstractmethod
    def read(self) -> T:
        pass


class DurinSensor(RunnableConsumer, Sensor[Observation]):

    _BUFFER_LENGTH = 50

    def __init__(self, link: UDPLink):
        self.link = link
        context = multiprocessing.get_context("spawn")

        self.tof = context.Array("f", 8 * 8 * 8)
        self.tof_ringbuffer = context.Array("d", self._BUFFER_LENGTH)
        self.tof_ringbuffer_idx = context.Value("i", 0)

        self.charge = context.Value("f", 0)
        self.voltage = context.Value("f", 0)
        self.imu = context.Array("f", 3 * 3)
        self.uwb = context.Array("f", 10*2)
        self.position = context.Array("f", 3)
        self.ringbuffer = context.Array("d", 100)
        self.ringbuffer_idx = context.Value("i", 0)
        self.timestamp_update = context.Value("d", time.time())
        ###
        self.status_ringbuffer = context.Array("d", self._BUFFER_LENGTH)
        self.status_ringbuffer_idx = context.Value("i", 0)

        self.imu_ringbuffer = context.Array("d", self._BUFFER_LENGTH)
        self.imu_ringbuffer_idx = context.Value("i", 0)

        self.uwb_ringbuffer = context.Array("d", self._BUFFER_LENGTH)
        self.uwb_ringbuffer_idx = context.Value("i", 0)

        super().__init__(
            self.link.buffer,
            self.tof,
            self.tof_ringbuffer,
            self.tof_ringbuffer_idx,
            self.charge,
            self.voltage,
            self.status_ringbuffer,
            self.status_ringbuffer_idx,
            self.imu,
            self.imu_ringbuffer,
            self.imu_ringbuffer_idx,
            self.uwb,
            self.uwb_ringbuffer,
            self.uwb_ringbuffer_idx,
            self.position,

        )

    def _read_buffer_frequency(self, buffer, index):
        buffer = RingBuffer(buffer.get_obj(), index.value)
        difference = buffer.get_newest() - buffer.get_oldest()
        if difference != 0:
            return self._BUFFER_LENGTH / difference

    def read(self):
        tof = np.frombuffer(self.tof.get_obj(),dtype=np.float32).reshape((8, 8, 8))
        imu = np.frombuffer(self.imu.get_obj(),dtype=np.float32).reshape((3, 3))
        uwb = np.frombuffer(self.uwb.get_obj(),dtype=np.float32).reshape((10,2))
        position = np.frombuffer(self.position.get_obj(),dtype=np.float32)
        
        return Observation(
            tof=tof,
            tof_frequency=self._read_buffer_frequency(self.tof_ringbuffer, self.tof_ringbuffer_idx),
            charge=self.charge.value,
            voltage=self.voltage.value,
            status_frequency=self._read_buffer_frequency(self.status_ringbuffer, self.status_ringbuffer_idx),
            imu=imu,
            uwb = uwb,
            position=position,
            imu_frequency=self._read_buffer_frequency(self.imu_ringbuffer, self.imu_ringbuffer_idx),
            uwb_frequency=self._read_buffer_frequency(self.uwb_ringbuffer, self.uwb_ringbuffer_idx),
        )

    def start(self):
        super().start()
        self.link.start()

    def stop(self):
        super().stop()
        self.link.stop()

    def consume(
        self,
        item,
        tof,
        tof_ringbuffer,
        tof_ringbuffer_idx,
        charge,
        voltage,
        status_ringbuffer,
        status_ringbuffer_idx,
        imu,
        imu_ringbuffer,
        imu_ringbuffer_idx,
        uwb,
        uwb_ringbuffer,
        uwb_ringbuffer_idx,
        position,

    ):
        which = item.which()

        try:
            if which == "tofObservations":
                for obs in item.tofObservations.observations:
                    data = np.array(obs.ranges)
                    tof.get_obj()[obs.id * 64: (obs.id+1) * 64] = data
                self._update_frequency(tof_ringbuffer, tof_ringbuffer_idx)


            elif which == "systemStatus":
                voltage.value = item.systemStatus.batteryMv
                charge.value = item.systemStatus.batteryPercent
                self._update_frequency(status_ringbuffer, status_ringbuffer_idx)


            elif which == "imuMeasurement":
                imu_quantity = [item.imuMeasurement.accelerometerX, item.imuMeasurement.accelerometerY, item.imuMeasurement.accelerometerZ,
                            item.imuMeasurement.gyroscopeX, item.imuMeasurement.gyroscopeY, item.imuMeasurement.gyroscopeZ,
                            item.imuMeasurement.magnetometerX, item.imuMeasurement.magnetometerY, item.imuMeasurement.magnetometerZ]
                for i in range(9):
                    imu.get_obj()[i] = imu_quantity[i]
                self._update_frequency(imu_ringbuffer, imu_ringbuffer_idx)


            
            elif which == "uwbNodes":
                node_list = item.uwbNodes.nodes

                uwb.get_obj()[:] = 20*[0]   # Reset beacon list

                for i in range(len(node_list)):
                    data = np.array([node_list[i].nodeId, node_list[i].distanceMm])
                    uwb.get_obj()[i*2:(i*2)+2] = data
                self._update_frequency(uwb_ringbuffer, uwb_ringbuffer_idx)

            try:
                if which == "position":
                    data = np.array([item.position.vectorMm.x,item.position.vectorMm.y,item.position.vectorMm.z])
                    position.get_obj()[:] = data
            except:
                pass
                
        except Exception as e:
            logging.warning("Error when receiving sensor data " + str(e))

    def _update_frequency(self, buffer_name, buffer_idx):
        buffer_obj = buffer_name.get_obj()
        buffer = RingBuffer(np.frombuffer(buffer_obj), buffer_idx.value)
        buffer.append(time.time())
        buffer_obj[:] = buffer.buffer
        buffer_idx.value = buffer.counter
        


import pygame
import math
import os
import random
import sys
import ctypes
import time

import numpy as np
from durin.actuator import Move

from durin.durin import Durin
from durin.io.gamepad import Gamepad
import durin


# Constants
surface_width = 200
surface_height = 200
sleep_interval = 0.02

SENSOR_PLACEMENTS = [
    (0.45, 0.4),
    (0.41, 0.03),
    (0.25, 0.02),
    (0.41, 0.7),
    (0.25, 0.78),
    (0.02, 0.7),
    (0.02, 0.03),
    (0.02, 0.4),
]

SENSOR_ROTATIONS = [180, 180+45, -90, 180-45, 90, 45, -45, 0]

# A distance (in % of screen size) constant related to the layout.
d= 0.02
x=0.68

TITLE_PLACEMENT = (x, 0.1)

IP_PLACEMENT = (x, 0.1 + 3* d)

BATTERY_PLACEMENT = (x,0.1 + 8*d)

IMU_PLACEMENT = (x+2*d, 0.1 +14*d)  # Upper left corner

IMU_INTEG_PLACEMENT = (x, 0.1 + 19*d)

POSITION_PLACEMENT = [(x,0.1+26*d), (x+2*d,0.1+26*d), (x+ 4*d, 0.1+26*d)]

UWB_PLACEMENT = (x, 0.1+32*d)              # Upper left corner





class DurinUI(Durin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gamepad = Gamepad()

        self.vertical = 0
        self.horizontal = 0
        self.tau = 0.9999
        self.rot = 0

    def __enter__(self):
        self.a = 0 # Just for debugging. Delete soon!

        pygame.init()
        self.clock = pygame.time.Clock()
        self.font = pygame.font.SysFont(None, 24)
        self.big_font = pygame.font.SysFont(None, 34)


        # Set up the display
        # Get screen size
        info = pygame.display.Info()
        self.screen_width, self.screen_height = info.current_w, info.current_h-100

        self.screen = pygame.display.set_mode((0, 0), pygame.RESIZABLE)

        # Buffer screen:
        #self.back_buffer = pygame.Surface((self.screen_width, self.screen_height))

        # Make it fullscreen
        if sys.platform == "win32":
            HWND = pygame.display.get_wm_info()['window']
            SW_MAXIMIZE = 3
            ctypes.windll.user32.ShowWindow(HWND, SW_MAXIMIZE)
        
        # Durin Image
        resource_file = "durin\durin\durin_birdseye.jpg"
        resource_path = os.path.join(os.getcwd(), resource_file)
        self.image = pygame.image.load(resource_path)
        self.image = pygame.transform.scale(self.image, (1.75*self.screen_width//3, self.screen_height))


        self.image_surface = pygame.Surface(self.image.get_size())
        self.image_surface.blit(self.image, (0,0))

        # Create surfaces for ToF data.
        self.surfaces = []
        for o in range(8):
            surface = pygame.Surface((surface_width, surface_height), pygame.SRCALPHA)
            self.surfaces.append(surface)

        pygame.display.update()




        return super().__enter__()
    
    def __exit__(self, e, b, t):
        pygame.quit()
        self.gamepad.stop()
        exit()
        return super().__exit__(e, b, t)
    
    def read_user_input(self, allow_movement: bool = True, sleep_interval: float=0.02):
        keys = pygame.key.get_pressed()

        # Gamepad
        if not self.gamepad.queue.empty():
            x, y, r = self.gamepad.queue.get()
            self.horizontal = x
            self.vertical = y
            self.rot = r
        
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                return False
            
            # Keyboard
            elif event.type == pygame.KEYDOWN:
                # Key pressed
                if event.key == pygame.K_UP or event.key == pygame.K_w:
                    self.vertical = 500
                elif event.key == pygame.K_DOWN or event.key == pygame.K_s:
                    self.vertical = -500
                elif event.key == pygame.K_LEFT or event.key == pygame.K_d:
                    self.horizontal = -500
                elif event.key == pygame.K_RIGHT or event.key == pygame.K_a:
                    self.horizontal = 500

            elif event.type == pygame.KEYUP:
                # Key released
                if event.key == pygame.K_UP or event.key == pygame.K_w or event.key == pygame.K_DOWN or event.key == pygame.K_s:
                    self.vertical = 0
                elif event.key == pygame.K_LEFT or event.key == pygame.K_d or event.key == pygame.K_RIGHT or event.key == pygame.K_a:
                    self.horizontal = 0

        if allow_movement:
            self(Move(self.horizontal, self.vertical, self.rot))

        time.sleep(sleep_interval) # Sleep to avoid sending too many commands

        return True
        
    
    def render_sensors(self, obs, size: int = 180):

        self.screen.fill((0,0,0))   # Fill screen with black
        self.screen.blit(self.image_surface, (0,0))

        
        # Update ToF-sensors ######################
        tofs = (np.tanh((obs.tof / 1000)) * 255).astype(np.int32)

        # Rotated surfaces
        rotated_surfaces = []

        for o in range(len(self.surfaces)):
            surface = self.surfaces[o]

            square_size = surface_width//8
            for i in range(8):
                for j in range(8):

                    left = i * square_size
                    top = j * square_size
                    square_rect = pygame.Rect(left, top, square_size, square_size)

                    color_value = tofs[o][i][j]
                    color = (color_value, color_value, color_value)
                    pygame.draw.rect(surface, color, square_rect)


            rotation_angle = SENSOR_ROTATIONS[o]
            rotated_surface = pygame.transform.rotate(surface, rotation_angle)
            rotated_surfaces.append(rotated_surface)


        for i in range(8):
            self.screen.blit(rotated_surfaces[i], (SENSOR_PLACEMENTS[i][0]*self.screen_width,SENSOR_PLACEMENTS[i][1]*self.screen_height))
                
        # Update UWB ######################

        uwb = obs.uwb
        #self.render_text("Becon ID\t\t\tDistance (mm)", UWB_PLACEMENT[0])


        for i in range(10):
            if uwb[i][0] != 0:
                self.render_text(str(uwb[i][0]), (UWB_PLACEMENT[0], UWB_PLACEMENT[1] + i*d))
                self.render_text(str(uwb[i][1]), (UWB_PLACEMENT[0]+ 5*d, UWB_PLACEMENT[1]+i*d))
            else:
                break



        # Update IMU ######################
        imu = obs.imu
        #type = ["Acce", "Gyro", "Magn."]
        for type in range(3):
            for xyz in range(3):
                self.render_text(str(imu[type][xyz]), (IMU_PLACEMENT[0]+xyz*3*d, IMU_PLACEMENT[1]+type*d))
        

        # Update battery level and voltage ######################
        voltage = obs.voltage
        charge = obs.charge
        self.render_text(str(charge) + " %", BATTERY_PLACEMENT)
        self.render_text(str(voltage) + " mV", (BATTERY_PLACEMENT[0]+5*d,BATTERY_PLACEMENT[1]))


        # Update Durin position ######################
        for m in range(3):
            self.render_text(str(obs.position[m]), POSITION_PLACEMENT[m])

        
        self.render_static_texts()

        # Just for debugging.
        self.a += 1
        self.render_text("Time step (for debugging): " + str(self.a),(UWB_PLACEMENT[0],UWB_PLACEMENT[1]+10*d))

        # Update screen
        pygame.display.update()

        #self.clock.tick(100)

    
    def render_text(self, input_text, position, color="w", size = "small"):
        if color == "w":
            c = (255,255,255)
        elif color == "o":
            c = (255,183,91)
        elif color == "b":
            c = (100,100,255)
        elif color == "t":
            c = (255,143,161)

        if size == "small":
            text =  self.font.render(input_text, True, c)
        elif size == "big":
            text=  self.big_font.render(input_text, True, c)
        self.screen.blit(text, (position[0]*self.screen_width,position[1]*self.screen_height))

    def render_static_texts(self):
        # Static texts¨

        self.render_text("Durin Dashboard", TITLE_PLACEMENT, "t", "big")

        self.render_text("IP address", IP_PLACEMENT, "o")
        self.render_text("MAC address", (IP_PLACEMENT[0]+5*d,IP_PLACEMENT[1]), "o")
        self.render_text("Durin ID", (IP_PLACEMENT[0]+10*d,IP_PLACEMENT[1]), "o")

        self.render_text("Integrated IMU data", (IMU_INTEG_PLACEMENT), "o")



        self.render_text("UWB ID", (UWB_PLACEMENT[0],UWB_PLACEMENT[1]-2*d), "o")
        self.render_text("Distance (mm)", (UWB_PLACEMENT[0]+5*d,UWB_PLACEMENT[1]-2*d), "o")

        self.render_text("IMU data",(IMU_PLACEMENT[0]-2*d,IMU_PLACEMENT[1]-3*d), "o")
        self.render_text("x",(IMU_PLACEMENT[0],IMU_PLACEMENT[1]-d), "b")
        self.render_text("y",(IMU_PLACEMENT[0]+3*d,IMU_PLACEMENT[1]-d),"b")
        self.render_text("z",(IMU_PLACEMENT[0]+6*d,IMU_PLACEMENT[1]-d),"b")
        self.render_text("Acce",(IMU_PLACEMENT[0]-2*d,IMU_PLACEMENT[1]),"b")
        self.render_text("Gyro",(IMU_PLACEMENT[0]-2*d,IMU_PLACEMENT[1]+d),"b")
        self.render_text("Magn",(IMU_PLACEMENT[0]-2*d,IMU_PLACEMENT[1]+2*d),"b")

        self.render_text("Battery level",(BATTERY_PLACEMENT[0],BATTERY_PLACEMENT[1]-d), "o")
        self.render_text("Voltage",(BATTERY_PLACEMENT[0]+5*d,BATTERY_PLACEMENT[1]-d), "o")


        self.render_text("Durin coordinates",(POSITION_PLACEMENT[0][0],POSITION_PLACEMENT[0][1]-3*d), "o")
        self.render_text("x",(POSITION_PLACEMENT[0][0],POSITION_PLACEMENT[0][1]-d),"b")
        self.render_text("y",(POSITION_PLACEMENT[1][0],POSITION_PLACEMENT[0][1]-d),"b")
        self.render_text("z",(POSITION_PLACEMENT[2][0],POSITION_PLACEMENT[0][1]-d),"b")
    




